@@ -1,20 +1,19 @@
 # -*- coding: utf-8 -*-
+# cython: language_level=3
 
-## Copyright 1999-2008 by LivingLogic AG, Bayreuth/Germany
-## Copyright 1999-2008 by Walter Dörwald
+## Copyright 1999-2014 by LivingLogic AG, Bayreuth/Germany
+## Copyright 1999-2014 by Walter Dörwald
 ##
 ## All Rights Reserved
 ##
-## See xist/__init__.py for the license
+## See ll/xist/__init__.py for the license
 
 
 """
 This module contains functions related to the handling of CSS.
 """
 
-from __future__ import with_statement
-
-import os, contextlib
+import os, contextlib, operator
 
 try:
 	import cssutils
@@ -23,7 +22,7 @@ except ImportError:
 	cssutils = None
 else:
 	import logging
-	cssutils.log.setloglevel(logging.FATAL)
+	cssutils.log.setLevel(logging.FATAL)
 
 from ll import misc, url
 from ll.xist import xsc, xfind
@@ -36,38 +35,33 @@ __docformat__ = "reStructuredText"
 def _isstyle(path):
 	if path:
 		node = path[-1]
-		return (isinstance(node, html.style) and unicode(node.attrs["type"]) == "text/css") or (isinstance(node, html.link) and unicode(node.attrs["rel"]) == "stylesheet")
+		return isinstance(node, (html.style, html.link)) and str(node.attrs.type) == "text/css"
 	return False
 
 
 def replaceurls(stylesheet, replacer):
 	"""
-	Replace all URLs appearing in the :class:`CSSStyleSheet` :var:`stylesheet`.
-	For each URL the function :var:`replacer` will be called and the URL will
+	Replace all URLs appearing in the :class:`CSSStyleSheet` :obj:`stylesheet`.
+	For each URL the function :obj:`replacer` will be called and the URL will
 	be replaced with the result.
 	"""
 	def newreplacer(u):
-		return unicode(replacer(url.URL(u)))
-	stylesheet.replaceUrls(newreplacer) # This needs at least r242 of cssutils
+		return str(replacer(url.URL(u)))
+	cssutils.replaceUrls(stylesheet, newreplacer)
 
 
 def geturls(stylesheet):
 	"""
 	Return a list of all URLs appearing in the :class:`CSSStyleSheet`
-	:var:`stylesheet`.
+	:obj:`stylesheet`.
 	"""
-	urls = []
-	def collector(u):
-		urls.append(url.URL(u))
-		return u
-	stylesheet.replaceUrls(collector)
-	return urls
+	return [url.URL(u) for u in cssutils.getUrls(stylesheet)] # This requires cssutils 0.9.5b1
 
 
 def _getmedia(stylesheet):
 	while stylesheet is not None:
 		if stylesheet.media is not None:
-			return set(mq.mediaType for mq in stylesheet.media)
+			return {mq.mediaType for mq in stylesheet.media}
 		stylesheet = stylesheet.parentStyleSheet
 	return None
 
@@ -92,106 +86,123 @@ def _doimport(wantmedia, parentsheet, base):
 					text = r.read()
 				sheet = css.CSSStyleSheet(href=str(href), media=havemedia, parentStyleSheet=parentsheet)
 				sheet.cssText = text
-				for rule in _doimport(wantmedia, sheet, href):
-					yield rule
+				yield from _doimport(wantmedia, sheet, href)
 			elif rule.type == css.CSSRule.MEDIA_RULE:
 				if wantmedia in (mq.mediaType for mq in rule.media):
-					for subrule in rule.cssRules:
-						yield subrule
+					yield from rule.cssRules
 			elif rule.type == css.CSSRule.STYLE_RULE:
 				yield rule
 
 
-def iterrules(node, base=None, media=None):
+def iterrules(node, base=None, media=None, title=None):
 	"""
-	Return an iterator for all CSS rules defined in the HTML tree :var:`node`.
+	Return an iterator for all CSS rules defined in the HTML tree :obj:`node`.
 	This will parse the CSS defined in any :class:`html.style` or
 	:class:`html.link` element (and recursively in those stylesheets imported
 	via the ``@import`` rule). The rules will be returned as
 	:class:`CSSStyleRule` objects from the :mod:`cssutils` package (so this
 	requires :mod:`cssutils`).
 
-	The :var:`base` argument will be used as the base URL for parsing the
+	The :obj:`base` argument will be used as the base URL for parsing the
 	stylesheet references in the tree (so :const:`None` means the URLs will be
 	used exactly as they appear in the tree). All URLs in the style properties
 	will be resolved.
 
-	If :var:`media` is given, only rules that apply to this media type will
+	If :obj:`media` is given, only rules that apply to this media type will
 	be produced.
+
+	:obj:`title` can be used to specify which stylesheet group should be used.
+	If :obj:`title` is :const:`None` only the persistent and preferred
+	stylesheets will be used. If :obj:`title` is a string only the persistent
+	stylesheets and alternate stylesheets with that style name will be used.
+
+	For a description of "persistent", "preferred" and "alternate" stylesheets
+	see <http://www.w3.org/TR/2002/WD-xhtml2-20020805/mod-styleSheet.html#sec_20.1.2.>
 	"""
 	if base is not None:
 		base = url.URL(base)
 
-	def doiter(node, base, media):
-		for cssnode in node.walknode(_isstyle):
+	def matchlink(node):
+		if node.attrs.media.hasmedia(media):
+			if title is None:
+				if "title" not in node.attrs and "alternate" not in str(node.attrs.rel).split():
+					return True
+			elif not node.attrs.title.isfancy() and str(node.attrs.title) == title and "alternate" in str(node.attrs.rel).split():
+				return True
+		return False
+
+	def matchstyle(node):
+		if node.attrs.media.hasmedia(media):
+			if title is None:
+				if "title" not in node.attrs:
+					return True
+			elif str(node.attrs.title) == title:
+				return True
+		return False
+
+	def doiter(node):
+		for cssnode in node.walknodes(_isstyle):
 			if isinstance(cssnode, html.style):
-				href = str(self.base) if base is not None else None
-				if cssnode.attrs.media.hasmedia(media):
-					stylesheet = cssutils.parseString(unicode(cssnode.content), href=href, media=unicode(cssnode.attrs.media))
-					for rule in _doimport(media, stylesheet, base):
-						yield rule
+				href = str(base) if base is not None else None
+				if matchstyle(cssnode):
+					stylesheet = cssutils.parseString(str(cssnode.content), href=href, media=str(cssnode.attrs.media))
+					yield from _doimport(media, stylesheet, base)
 			else: # link
 				if "href" in cssnode.attrs:
-					href = cssnode.attrs["href"].asURL()
+					href = cssnode.attrs.href.asURL()
 					if base is not None:
 						href = base/href
-					if cssnode.attrs.media.hasmedia(media):
-						with contextlib.closing(href.open("rb")) as r:
-							s = r.read()
-						stylesheet = cssutils.parseString(unicode(s), href=str(href), media=unicode(cssnode.attrs.media))
-						for rule in _doimport(media, stylesheet, href):
-							yield rule
-	return misc.Iterator(doiter(node, base, media))
+					if matchlink(cssnode):
+						stylesheet = cssutils.parseUrl(str(href), media=str(cssnode.attrs.media))
+						yield from _doimport(media, stylesheet, href)
+	return misc.Iterator(doiter(node))
 
 
-def applystylesheets(node, base=None, media=None):
+def applystylesheets(node, base=None, media=None, title=None):
 	"""
-	:func:`applystylesheets` modifies the XIST tree :var:`node` by removing all
+	:func:`applystylesheets` modifies the XIST tree :obj:`node` by removing all
 	CSS (from :class:`html.link` and :class:`html.style` elements and their
-	``@import``ed stylesheets) and puts the resulting styles properties into
-	the ``style`` attribute of the every affected element instead.
-	
-	The :var:`base` argument will be used as the base URL for parsing the
-	stylesheet references in the tree (so :const:`None` means the URLs will be
-	used exactly as they appear in the tree). All URLs in the style properties
-	will be resolved.
+	``@import``\ed stylesheets) and puts the resulting style properties into
+	the ``style`` attribute of every affected element instead.
 
-	If :var:`media` is given, only rules that apply to this media type will
-	be applied.
+	For the meaning of :obj:`base`, :obj:`media` and :obj:`title` see
+	:func:`iterrules`.
 	"""
+
 	def iterstyles(node, rules):
-		for data in rules:
-			yield data
+		yield from rules
 		# According to CSS 2.1 (http://www.w3.org/TR/CSS21/cascade.html#specificity)
 		# style attributes have the highest weight, so we yield it last
 		# (CSS 3 uses the same weight)
 		if "style" in node.attrs:
-			style = node.attrs["style"]
+			style = node.attrs.style
 			if not style.isfancy():
 				yield (
+					(1, 0, 0, 0),
 					xfind.IsSelector(node),
-					cssutils.parseString(u"*{%s}" % style).cssRules[0].style # parse the style out of the style attribute
+					cssutils.parseStyle(str(style)) # parse the style out of the style attribute
 				)
 
 	rules = []
-	for (i, rule) in enumerate(iterrules(node, base=base, media=media)):
+	for rule in iterrules(node, base=base, media=media, title=title):
 		for sel in rule.selectorList:
-			rules.append((selector(sel), rule.style))
-	rules.sort(key=lambda(sel, rule): sel.cssweight())
+			rules.append((sel.specificity, selector(sel), rule.style))
+	rules.sort(key=operator.itemgetter(0))
 	count = 0
-	for path in node.walk(xsc.Element):
-		del path[-1][_isstyle] # drop style sheet nodes
-		if path[-1].Attrs.isallowed("style"):
+	for cursor in node.walk(xsc.Element):
+		del cursor.node[_isstyle] # drop style sheet nodes
+		if cursor.node.Attrs.isdeclared("style"):
 			styles = {}
-			for (sel, style) in iterstyles(path[-1], rules):
-				if sel.matchpath(path):
-					for prop in style.seq:
-						if not isinstance(prop, css.CSSComment):
-							styles[prop.name] = (count, prop.name, prop.cssValue.cssText)
-							count += 1
-			style = " ".join("%s: %s;" % (name, value) for (count, name, value) in sorted(styles.itervalues()))
+			for (spec, sel, style) in iterstyles(cursor.node, rules):
+				if cursor.path in sel:
+					for prop in style:
+						# Properties from later rules overwrite those from earlier ones
+						# We're storing the count so that sorting keeps the order
+						styles[prop.name] = (count, prop.cssText)
+						count += 1
+			style = " ".join("{};".format(value) for (count, value) in sorted(styles.values()))
 			if style:
-				path[-1].attrs["style"] = style
+				cursor.node.attrs.style = style
 
 
 ###
@@ -199,8 +210,8 @@ def applystylesheets(node, base=None, media=None):
 ###
 
 def _is_nth_node(iterator, node, index):
-	# Return whether :var:`node` is the :var:`index`'th node in :var:`iterator` (starting at 1)
-	# index is an int or int string or "even" or "odd"
+	# Return whether :obj:`node` is the :obj:`index`'th node in :obj:`iterator` (starting at 1)
+	# :obj:`index` is an int or int string or "even" or "odd"
 	if index == "even":
 		for (i, child) in enumerate(iterator):
 			if child is node:
@@ -212,11 +223,11 @@ def _is_nth_node(iterator, node, index):
 				return i % 2 == 0
 		return False
 	else:
-		if not isinstance(index, (int, long)):
+		if not isinstance(index, int):
 			try:
 				index = int(index)
 			except ValueError:
-				raise ValueError("illegal argument %r" % index)
+				raise ValueError("illegal argument {!r}".format(index))
 			else:
 				if index < 1:
 					return False
@@ -227,8 +238,8 @@ def _is_nth_node(iterator, node, index):
 
 
 def _is_nth_last_node(iterator, node, index):
-	# Return whether :var:`node` is the :var:`index`'th last node in :var:`iterator`
-	# index is an int or int string or "even" or "odd"
+	# Return whether :obj:`node` is the :obj:`index`'th last node in :obj:`iterator`
+	# :obj:`index` is an int or int string or "even" or "odd"
 	if index == "even":
 		pos = None
 		for (i, child) in enumerate(iterator):
@@ -242,11 +253,11 @@ def _is_nth_last_node(iterator, node, index):
 				pos = i
 		return pos is None or (i-pos) % 2 == 0
 	else:
-		if not isinstance(index, (int, long)):
+		if not isinstance(index, int):
 			try:
 				index = int(index)
 			except ValueError:
-				raise ValueError("illegal argument %r" % index)
+				raise ValueError("illegal argument {!r}".format(index))
 			else:
 				if index < 1:
 					return False
@@ -270,8 +281,6 @@ class CSSWeightedSelector(xfind.Selector):
 	"""
 	Base class for all CSS pseudo-class selectors.
 	"""
-	def cssweight(self):
-		return xfind.CSSWeight(0, 1, 0)
 
 
 class CSSHasAttributeSelector(CSSWeightedSelector):
@@ -282,55 +291,69 @@ class CSSHasAttributeSelector(CSSWeightedSelector):
 	def __init__(self, attributename):
 		self.attributename = attributename
 
-	def matchpath(self, path):
+	def __contains__(self, path):
 		if path:
 			node = path[-1]
-			if isinstance(node, xsc.Element) and node.Attrs.isallowed_xml(self.attributename):
-				return node.attrs.has_xml(self.attributename)
+			if isinstance(node, xsc.Element):
+				return node.attrs.has(self.attributename)
 		return False
 
 	def __str__(self):
-		return "%s(%r)" % (self.__class__.__name__, self.attributename)
+		return "{0.__class___.__qualname__}({0.attributename!r})".format(self)
 
 
 class CSSAttributeListSelector(CSSWeightedSelector):
+	"""
+	A :class:`CSSAttributeListSelector` selector selects all element nodes
+	where an attribute with the specified XML name has the specified word
+	among the white space-separated list of words in the attribute value.
+	"""
 	def __init__(self, attributename, attributevalue):
 		self.attributename = attributename
 		self.attributevalue = attributevalue
 
-	def matchpath(self, path):
+	def __contains__(self, path):
 		if path:
 			node = path[-1]
-			if isinstance(node, xsc.Element) and node.Attrs.isallowed_xml(self.attributename):
-				attr = node.attrs.get_xml(self.attributename)
-				return self.attributevalue in unicode(attr).split()
+			if isinstance(node, xsc.Element):
+				attr = node.attrs.get(self.attributename)
+				return self.attributevalue in str(attr).split()
 		return False
 
 	def __str__(self):
-		return "%s(%r, %r)" % (self.__class__.__name__, self.attributename, self.attributevalue)
+		return "{0.__class__.__qualname__}({0.attributename!r}, {0.attributevalue!r})".format(self)
 
 
 class CSSAttributeLangSelector(CSSWeightedSelector):
+	"""
+	A :class:`CSSAttributeLangSelector` selector selects all element nodes
+	where an attribute with the specified XML name either is exactly the
+	specified value or starts with the specified value followed by ``"-"``.
+	"""
 	def __init__(self, attributename, attributevalue):
 		self.attributename = attributename
 		self.attributevalue = attributevalue
 
-	def matchpath(self, path):
+	def __contains__(self, path):
 		if path:
 			node = path[-1]
-			if isinstance(node, xsc.Element) and node.Attrs.isallowed_xml(self.attributename):
-				attr = node.attrs.get_xml(self.attributename)
-				parts = unicode(attr).split("-", 1)
+			if isinstance(node, xsc.Element):
+				attr = node.attrs.get(self.attributename)
+				parts = str(attr).split("-", 1)
 				if parts:
 					return parts[0] == self.attributevalue
 		return False
 
 	def __str__(self):
-		return "%s(%r, %r)" % (self.__class__.__name__, self.attributename, self.attributevalue)
+		return "{0.__class__.__qualname__}({0.attributename!r}, {0.attributevalue!r})".format(self)
 
 
 class CSSFirstChildSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSFirstChildSelector` selector selects all element nodes
+	that are the first child of its parent.
+	"""
+	def __contains__(self, path):
 		return len(path) >= 2 and _is_nth_node(path[-2][xsc.Element], path[-1], 1)
 
 	def __str__(self):
@@ -338,7 +361,11 @@ class CSSFirstChildSelector(CSSWeightedSelector):
 
 
 class CSSLastChildSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSLastChildSelector` selector selects all element nodes
+	that are the last child of its parent.
+	"""
+	def __contains__(self, path):
 		return len(path) >= 2 and _is_nth_last_node(path[-2][xsc.Element], path[-1], 1)
 
 	def __str__(self):
@@ -346,7 +373,11 @@ class CSSLastChildSelector(CSSWeightedSelector):
 
 
 class CSSFirstOfTypeSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSLastChildSelector` selector selects all element nodes
+	that are the first of its type among their siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			return isinstance(node, xsc.Element) and _is_nth_node(misc.Iterator(_children_of_type(path[-2], node.xmlname)), node, 1)
@@ -357,7 +388,11 @@ class CSSFirstOfTypeSelector(CSSWeightedSelector):
 
 
 class CSSLastOfTypeSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSLastChildSelector` selector selects all element nodes
+	that are the last of its type among their siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			return isinstance(node, xsc.Element) and _is_nth_last_node(misc.Iterator(_children_of_type(path[-2], node.xmlname)), node, 1)
@@ -368,7 +403,11 @@ class CSSLastOfTypeSelector(CSSWeightedSelector):
 
 
 class CSSOnlyChildSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSOnlyChildSelector` selector selects all element nodes that are
+	the only element among its siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -383,7 +422,11 @@ class CSSOnlyChildSelector(CSSWeightedSelector):
 
 
 class CSSOnlyOfTypeSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSOnlyOfTypeSelector` selector selects all element nodes that are
+	the only element of its type among its siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -398,7 +441,11 @@ class CSSOnlyOfTypeSelector(CSSWeightedSelector):
 
 
 class CSSEmptySelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSEmptySelector` selector selects all element nodes that are
+	empty (i.e. they contain no elements or non-whitespace text).
+	"""
+	def __contains__(self, path):
 		if path:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -413,7 +460,10 @@ class CSSEmptySelector(CSSWeightedSelector):
 
 
 class CSSRootSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSRootSelector` selector selects the root element.
+	"""
+	def __contains__(self, path):
 		return len(path) == 1 and isinstance(path[-1], xsc.Element)
 
 	def __str__(self):
@@ -421,22 +471,25 @@ class CSSRootSelector(CSSWeightedSelector):
 
 
 class CSSLinkSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSLinkSelector` selector selects all HTML links.
+	"""
+	def __contains__(self, path):
 		if path:
 			node = path[-1]
 			return isinstance(node, xsc.Element) and node.xmlns=="http://www.w3.org/1999/xhtml" and node.xmlname=="a" and "href" in node.attrs
 		return False
 
 	def __str__(self):
-		return "%s()" % self.__class__.__name__
+		return "{}()".format(self.__class__.__name__)
 
 
 class CSSInvalidPseudoSelector(CSSWeightedSelector):
-	def matchpath(self, path):
+	def __contains__(self, path):
 		return False
 
 	def __str__(self):
-		return "%s()" % self.__class__.__name__
+		return "{}()".format(self.__class__.__name__)
 
 
 class CSSHoverSelector(CSSInvalidPseudoSelector):
@@ -464,15 +517,22 @@ class CSSBeforeSelector(CSSInvalidPseudoSelector):
 
 
 class CSSFunctionSelector(CSSWeightedSelector):
+	"""
+	Base class of all CSS selectors that require an argument.
+	"""
 	def __init__(self, value=None):
 		self.value = value
 
 	def __str__(self):
-		return "%s(%r)" % (self.__class__.__name__, self.value)
+		return "{}({!r})".format(self.__class__.__name__, self.value)
 
 
 class CSSNthChildSelector(CSSFunctionSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSNthChildSelector` selector selects all element nodes that are
+	the n-th element among their siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -481,7 +541,11 @@ class CSSNthChildSelector(CSSFunctionSelector):
 
 
 class CSSNthLastChildSelector(CSSFunctionSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSNthLastChildSelector` selector selects all element nodes that are
+	the n-th last element among their siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -490,7 +554,11 @@ class CSSNthLastChildSelector(CSSFunctionSelector):
 
 
 class CSSNthOfTypeSelector(CSSFunctionSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSNthOfTypeSelector` selector selects all element nodes that are
+	the n-th of its type among their siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -499,7 +567,11 @@ class CSSNthOfTypeSelector(CSSFunctionSelector):
 
 
 class CSSNthLastOfTypeSelector(CSSFunctionSelector):
-	def matchpath(self, path):
+	"""
+	A :class:`CSSNthOfTypeSelector` selector selects all element nodes that are
+	the n-th last of its type among their siblings.
+	"""
+	def __contains__(self, path):
 		if len(path) >= 2:
 			node = path[-1]
 			if isinstance(node, xsc.Element):
@@ -513,13 +585,13 @@ class CSSTypeSelector(xfind.Selector):
 		self.xmlns = xsc.nsname(xmlns)
 		self.selectors = [] # id, class, attribute etc. selectors for this node
 
-	def matchpath(self, path):
+	def __contains__(self, path):
 		if path:
 			node = path[-1]
 			if self.type is None or node.xmlname == self.type:
 				if self.xmlns is None or node.xmlns == self.xmlns:
 					for selector in self.selectors:
-						if not selector.matchpath(path):
+						if path not in selector:
 							return False
 					return True
 		return False
@@ -537,22 +609,16 @@ class CSSTypeSelector(xfind.Selector):
 		v.append(")")
 		return "".join(v)
 
-	def cssweight(self):
-		result = xfind.CSSWeight(0, 0, 0, int(self.type is not None))
-		for selector in self.selectors:
-			result += selector.cssweight()
-		return result
-
 
 class CSSAdjacentSiblingCombinator(xfind.BinaryCombinator):
 	"""
-	A :class:`CSSAdjacentSiblingCombinator` work similar to an
-	:class:`AdjacentSiblingCombinator` except that only preceding elements
-	are considered.
+	A :class:`CSSAdjacentSiblingCombinator` works similar to an
+	:class:`xfind.AdjacentSiblingCombinator` except that only preceding
+	*elements* are considered.
 	"""
 
-	def matchpath(self, path):
-		if len(path) >= 2 and self.right.matchpath(path):
+	def __contains__(self, path):
+		if len(path) >= 2 and path in self.right:
 			# Find sibling
 			node = path[-1]
 			sibling = None
@@ -561,32 +627,32 @@ class CSSAdjacentSiblingCombinator(xfind.BinaryCombinator):
 					break
 				sibling = child
 			if sibling is not None:
-				return self.left.matchpath(path[:-1]+[sibling])
+				return path[:-1]+[sibling] in self.left
 		return False
 
 	def __str__(self):
-		return "%s(%s, %s)" % (self.__class__.__name__, self.left, self.right)
+		return "{}({}, {})".format(self.__class__.__name__, self.left, self.right)
 
 
 class CSSGeneralSiblingCombinator(xfind.BinaryCombinator):
 	"""
-	A :class:`CSSGeneralSiblingCombinator` work similar to an
-	:class:`GeneralSiblingCombinator` except that only preceding elements
+	A :class:`CSSGeneralSiblingCombinator` works similar to an
+	:class:`xfind.GeneralSiblingCombinator` except that only preceding *elements*
 	are considered.
 	"""
 
-	def matchpath(self, path):
-		if len(path) >= 2 and self.right.matchpath(path):
+	def __contains__(self, path):
+		if len(path) >= 2 and path in self.right:
 			node = path[-1]
 			for child in path[-2][xsc.Element]:
 				if child is node: # no previous element siblings
 					return False
-				if self.left.matchpath(path[:-1]+[child]):
+				if path[:-1]+[child] in self.left:
 					return True
 		return False
 
 	def __str__(self):
-		return "%s(%s, %s)" % (self.__class__.__name__, self.left, self.right)
+		return "{}({}, {})".format(self.__class__.__name__, self.left, self.right)
 
 
 _pseudoname2class = {
@@ -617,18 +683,18 @@ _function2class = {
 
 def selector(selectors, prefixes=None):
 	"""
-	Create a walk filter that will yield all nodes that match the specified
-	CSS expression. :var:`selectors` can be a string or a
-	:class:`cssutils.css.selector.Selector` object. :var:`prefixes`
+	Create a :class:`xfind.Selector` object that matches all nodes that match
+	the specified CSS selector expression. :obj:`selectors` can be a string or a
+	:class:`cssutils.css.selector.Selector` object. :obj:`prefixes`
 	may be a mapping mapping namespace prefixes to namespace names.
 	"""
 
-	if isinstance(selectors, basestring):
+	if isinstance(selectors, str):
 		if prefixes is not None:
-			prefixes = dict((key, xsc.nsname(value)) for (key, value) in prefixes.iteritems())
-			selectors = "%s\n%s{}" % ("\n".join("@namespace %s %r;" % (key if key is not None else "", value) for (key, value) in prefixes.iteritems()), selectors)
+			prefixes = dict((key, xsc.nsname(value)) for (key, value) in prefixes.items())
+			selectors = "{}\n{}{{}}".format("\n".join("@namespace {} {!r};".format(key if key is not None else "", value) for (key, value) in prefixes.items()), selectors)
 		else:
-			selectors = "%s{}" % selectors
+			selectors = "{}{{}}".format(selectors)
 		for rule in cssutils.CSSParser().parseString(selectors).cssRules:
 			if isinstance(rule, css.CSSStyleRule):
 				selectors = rule.selectorList.seq
@@ -640,19 +706,16 @@ def selector(selectors, prefixes=None):
 	elif isinstance(selectors, css.Selector):
 		selectors = [selectors]
 	else:
-		raise TypeError("can't handle %r" % type(selectors))
+		raise TypeError("can't handle {!r}".format(type(selectors)))
 	orcombinators = []
 	for selector in selectors:
 		rule = root = CSSTypeSelector()
-		prefix = None
 		attributename = None
 		attributevalue = None
 		combinator = None
-		inattr = False
 		for item in selector.seq:
 			t = item.type
 			v = item.value
-			print (t,v)
 			if t == "type-selector":
 				rule.xmlns = v[0] if v[0] != -1 else None
 				rule.type = v[1]
@@ -664,40 +727,38 @@ def selector(selectors, prefixes=None):
 			elif t == "class":
 				rule.selectors.append(xfind.hasclass(v.lstrip(".")))
 			elif t == "attribute-start":
-				inattr = True
 				combinator = None
 			elif t == "attribute-end":
 				if combinator is None:
 					rule.selectors.append(CSSHasAttributeSelector(attributename))
 				else:
 					rule.selectors.append(combinator(attributename, attributevalue))
-				inattr = False
 			elif t == "attribute-selector":
 				attributename = v
 			elif t == "equals":
-				combinator = xfind.attrhasvalue_xml
+				combinator = xfind.attrhasvalue
 			elif t == "includes":
 				combinator = CSSAttributeListSelector
 			elif t == "dashmatch":
 				combinator = CSSAttributeLangSelector
 			elif t == "prefixmatch":
-				combinator = xfind.attrstartswith_xml
+				combinator = xfind.attrstartswith
 			elif t == "suffixmatch":
-				combinator = xfind.attrendswith_xml
+				combinator = xfind.attrendswith
 			elif t == "substringmatch":
-				combinator = xfind.attrcontains_xml
+				combinator = xfind.attrcontains
 			elif t == "pseudo-class":
 				if v.endswith("("):
 					try:
 						rule.selectors.append(_function2class[v.lstrip(":").rstrip("(")]())
 					except KeyError:
-						raise ValueError("unknown function %s" % v)
+						raise ValueError("unknown function {}".format(v))
 					rule.function = v
 				else:
 					try:
 						rule.selectors.append(_pseudoname2class[v.lstrip(":")]())
 					except KeyError:
-						raise ValueError("unknown pseudo-class %s" % v)
+						raise ValueError("unknown pseudo-class {}".format(v))
 			elif t == "NUMBER":
 				# can only appear in a function => set the function value
 				rule.selectors[-1].value = v
@@ -721,23 +782,42 @@ def selector(selectors, prefixes=None):
 
 
 def parsestring(data, base=None, encoding=None):
+	"""
+	Parse the string :obj:`data` into a :mod:`cssutils` stylesheet. :obj:`base`
+	is the base URL for the parsing process, :obj:`encoding` can be used to force
+	the parser to use the specified encoding.
+	"""
 	if encoding is None:
 		encoding = "css"
-	stylesheet = cssutils.parseString(data.decode(encoding))
 	if base is not None:
 		base = url.URL(base)
+		href = str(base)
+	else:
+		href = None
+	stylesheet = cssutils.parseString(data.decode(encoding), href=href)
+	if base is not None:
 		def prependbase(u):
 			return base/u
 		replaceurls(stylesheet, prependbase)
-		stylesheet.href = str(base)
 	return stylesheet
 
 
 def parsestream(stream, base=None, encoding=None):
+	"""
+	Parse a :mod:`cssutils` stylesheet from the stream :obj:`stream`. :obj:`base`
+	is the base URL for the parsing process, :obj:`encoding` can be used to force
+	the parser to use the specified encoding.
+	"""
 	return parsestring(stream.read(), base=base, encoding=None)
 
 
 def parsefile(filename, base=None, encoding=None):
+	"""
+	Parse a :mod:`cssutils` stylesheet from the file named :obj:`filename`.
+	:obj:`base` is the base URL for the parsing process (defaulting to the
+	filename itself), :obj:`encoding` can be used to force the parser to use the
+	specified encoding.
+	"""
 	filename = os.path.expanduser(filename)
 	if base is None:
 		base = filename
@@ -746,6 +826,14 @@ def parsefile(filename, base=None, encoding=None):
 
 
 def parseurl(name, base=None, encoding=None, *args, **kwargs):
+	"""
+	Parse a :mod:`cssutils` stylesheet from the URL :obj:`name`. :obj:`base` is
+	the base URL for the parsing process (defaulting to the final URL of the
+	response, i.e. including redirects), :obj:`encoding` can be used to force
+	the parser to use the specified encoding. :obj:`arg` and :obj:`kwargs` are
+	passed on to :meth:`URL.openread`, so you can pass POST data and request
+	headers.
+	"""
 	with contextlib.closing(url.URL(name).openread(*args, **kwargs)) as stream:
 		if base is None:
 			base = stream.finalurl()
